/*
 Illinois Open Source License
 
 University of Illinois/NCSA
 Open Source License
 
 Copyright © 2009,    University of Illinois.  All rights reserved.
 
 Developed by: 
 Innovative Systems Lab
 National Center for Supercomputing Applications
 http://www.ncsa.uiuc.edu/AboutUs/Directorates/ISL.html
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal with the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimers.
 
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimers in the documentation and/or other materials provided with the distribution.
 
 * Neither the names of Innovative Systems Lab and National Center for Supercomputing Applications, nor the names of its contributors may be used to endorse or promote products derived from this Software without specific prior written permission.
 
 THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include <strings.h>

#include "model.h"
#include "nallatech.h"

#include "args.h"

#define TDIFF(ts, te) (te.tv_sec - ts.tv_sec + (te.tv_usec - ts.tv_usec) * 1e-6)

double *init_bins(int bins_per_dec, float min_angle, float max_angle, int angle_units, int *nbins);

void write_results(long long **DD, long long **RRS, long long **DRS, int njk, int nbins, int bins_per_dec, float min_angle, int random_count, char *fname);

void doCompute(struct accelerator *myacc, int co_proc,
               struct cartesian *data1, int n1, 
               struct cartesian *data2, int n2,
               int doSelf, 
               long long **data_bins, int nbins, double *binb, int njk);

int main (int argc, char *argv[])
{
    int rf, k;

    // parce cmd arguments
    options args;
    parse_args(argc, argv, &args);

    printf("\ndata file: %s \n", args.data_name);
    printf("number of data points: %d\n", args.ndpoints);
    //printf("data points per segment: %d\n", args.maxdsize);

    printf("random data files: %s.1-%i\n", args.random_name, args.random_count);
    printf("number of random points: %d \n", args.nrpoints);
    //printf("random points per segment: %d\n", args.maxrsize);
    //printf("compute RRS: %s\n", (args.doRRS) ? "YES" : "NO");

    printf("output file: %s \n", args.output_name);

    printf("njk: %i \n", args.njk);
    printf("co-processor: %s \n", (args.co_proc==CO_PROC_FPGA) ? "H101" : "none");

    int nbins;
    double *binb = init_bins(args.bins_per_dec, args.min_angle, args.max_angle, args.angle_units, &nbins);

    // memory for DD[njk]
    long long **DD = (long long **)malloc((args.njk+1)*sizeof(long long*));
    for (k = 0; k < args.njk+1; k++)
    {
        DD[k] = (long long*)malloc((nbins+2)*sizeof(long long));
        if (DD[k] == NULL) 
        {
            fprintf(stderr, "Unable to allocate memory for bin DD counts\n");
            exit(0);
        }
        bzero(DD[k], (nbins+2)*sizeof(long long));
    }

    // memory for RR
    long long **RRS = (long long **)malloc(1*sizeof(long long*));
    RRS[0] = (long long*)malloc((nbins+2)*sizeof(long long));
    if (RRS[0] == NULL) 
    {
        fprintf(stderr, "Unable to allocate memory for bin RRS counts\n");
        exit(0);
    }
    bzero(RRS[0], (nbins+2)*sizeof(long long));

    // memory for DR
    long long **DRS = (long long **)malloc((args.njk+1)*sizeof(long long*));
    for (k = 0; k < args.njk+1; k++)
    {
        DRS[k] = (long long*)malloc((nbins+2)*sizeof(long long));
        if (DRS[k] == NULL)
        {
            fprintf(stderr, "Unable to allocate memory for bin DRS counts\n");
            exit(0);
        }
        bzero(DRS[k], (nbins+2)*sizeof(long long));
    }

    // memory for input data
    struct cartesian *data = (struct cartesian*)malloc(args.ndpoints*sizeof(struct cartesian));
    if (data == NULL)
    {
        fprintf(stderr, "Unable to allocate memory for % data points (#1)\n", args.ndpoints);
        exit(0);
    }

    struct cartesian *random = (struct cartesian*)malloc(args.nrpoints*sizeof(struct cartesian));
    if (random == NULL)
    {
        fprintf(stderr, "Unable to allocate memory for % data points (#2)\n", args.nrpoints);
        exit(0);
    }

    // 
    struct accelerator *myacc = NULL;
    if (args.co_proc == CO_PROC_FPGA)
    {
        printf("Attempting to allocate H101 board...\n");
        myacc = nallatech_init();
        if (myacc == NULL)
        {
            printf("Failed to allocate H101 board, switching to no-coprocessor mode\n");
            args.co_proc = CO_PROC_CPU;
        }
        else
        {
            printf("Allocated H101 board, continuing in co-processor mode\n");
        }
    }

    struct timeval t2, t1, t0;
    float t_computeDD=0, t_computeRRS=0, t_computeDRS=0, t_fileIO=0;

    int npd, npr;

    // loop over all files
    for (rf = 0; rf < args.random_count+1; rf++)
    {
        printf("Processing dataset # %i...\n", rf);

        gettimeofday(&t0, NULL);

        if (rf == 0)
        {
            npd = readdatafile(args.data_name, data, args.ndpoints);
            if (npd != args.ndpoints)
            {
                fprintf(stderr, "Error: read %i data points out of %i\n", npd, args.ndpoints);
                exit(0);
            }

            assign_jk(data, npd, args.njk);
        }
        else
        {
            char fname[256];
            sprintf(fname, "%s.%i\0", args.random_name, rf);

            npr = readdatafile(fname, random, args.nrpoints);
            if (npr != args.nrpoints)
            {
                fprintf(stderr, "Error: read %i random points out of %i in file %s\n", npr, args.nrpoints, fname);
                exit(0);
            }
        }

        gettimeofday(&t1, NULL);
        t_fileIO += TDIFF(t0, t1);

        if (rf == 0)
        {
            gettimeofday(&t0, NULL);
            doCompute(myacc, args.co_proc, data, npd, data, npd, 1, DD, nbins, binb, args.njk+1);
            gettimeofday(&t1, NULL);

            t_computeDD += TDIFF(t0, t1);
        }
        else
        {
            gettimeofday(&t0, NULL);
            doCompute(myacc, args.co_proc, random, npr, random, npr, 1, RRS, nbins, binb, 1);
            gettimeofday(&t1, NULL);
            doCompute(myacc, args.co_proc, data, npd, random, npr, 0, DRS, nbins, binb, args.njk+1);
            gettimeofday(&t2, NULL);

            t_computeRRS += TDIFF(t0, t1);
            t_computeDRS += TDIFF(t1, t2);
        }
    }

    if (args.co_proc == CO_PROC_FPGA)
        nallatech_close(myacc);

    write_results(DD, RRS, DRS, args.njk, nbins, args.bins_per_dec, args.min_angle, args.random_count, args.output_name);

    printf("================================================\n");
    printf("Time to compute DD: %.4f sec\n", t_computeDD);
    printf("Time to compute RRS: %.4f sec\n", t_computeRRS);
    printf("Time to compute DRS: %.4f sec\n", t_computeDRS);
    printf("Time to load data files: %.4f sec\n", t_fileIO);
    printf("Time to compute DD, RRS, & DRS: %.4f sec\n", t_computeDD+t_computeRRS+t_computeDRS);
    printf("TOTAL time (DD+RRS+DRS+IO): %.4f sec\n", t_computeDD+t_computeRRS+t_computeDRS+t_fileIO);
    printf("================================================\n");

    // free memory
    free((char *)data);
    free((char *)random);
    free((char *)binb);

    for (k = 0; k < args.njk+1; k++)
    {
        free((char *)DD[k]);
        free((char *)DRS[k]);
    }

    free((char *)RRS[0]);

    free((char *)DD);
    free((char *)RRS);
    free((char *)DRS);

    return 0;
}


void write_results(long long **DD, long long **RRS, long long **DRS, int njk, int nbins, int bins_per_dec, float min_angle, int random_count, char *fname)
{
    // compute and output results
    FILE *outfile;
    int k, l;

    if ((outfile = fopen(fname, "w")) == NULL)
    {
        fprintf(stderr, "Unable to open output file %s for writing, assuming stdout\n", fname);
        outfile = stdout;
    }

    fprintf(outfile, "boundary DD DRS RRS w ");

    for (l = 1; l < njk+1; l++)
        fprintf(outfile, "DD[%i] DRS[%i] RRS w[%i]  ", l, l, l);

    fprintf(outfile, "\n");

    double bb = pow(10, log10(min_angle) + 0*1.0/bins_per_dec);
    fprintf(outfile, "<%f %lld %lld %lld -\n", bb, DD[0][0], DRS[0][0], RRS[0][0]);

    for (k = 1; k < nbins+1; k++)
    {
        double bb = pow(10, log10(min_angle) + k*1.0/bins_per_dec);
        fprintf(outfile, "%f ", bb);

        for (l = 0; l < njk+1; l++)
        {
          if (RRS[0][k] != 0)
          {
            double norm_dd = (double)DD[l][k]; // (args.ndpoints * (args.ndpoints - 1) / 2.0);
            double norm_rrs = (double)RRS[0][k]; // (args.random_count * args.nrpoints * (args.nrpoints - 1) / 2.0);
            double norm_drs = (double)DRS[l][k]; // (args.random_count * args.ndpoints * args.nrpoints);
            double w = (random_count * norm_dd - norm_drs) / norm_rrs + 1.0;

            fprintf(outfile, "%lld %lld %lld %f ", DD[l][k], DRS[l][k], RRS[0][k], w);
          }
          else
            fprintf(outfile, "%lld %lld %lld - ", DD[l][k], DRS[l][k], RRS[0][k]);
        }

        fprintf(outfile, "\n");
    }

    bb = pow(10, log10(min_angle) + nbins*1.0/bins_per_dec);
    fprintf(outfile, ">%f %lld %lld %lld\n", bb, DD[0][nbins+1], DRS[0][nbins+1], RRS[0][nbins+1]);

    fclose(outfile);
}


double *init_bins(int bins_per_dec, float min_angle, float max_angle, int angle_units, int *nbins)
{
    int k;

    *nbins = floor(bins_per_dec * (log10(max_angle) - log10(min_angle)));

    // memory for bin boundaries
    double *binb = (double *)malloc(((*nbins)+2)*sizeof(long long));
    if (binb == NULL)
    {
        fprintf(stderr, "Unable to allocate memory for bin boundaries\n");
        exit(0);
    }

    printf("Min angualr distance: %f %s\n", min_angle, (angle_units) ? "arcmin" : "degrees");
    printf("Max angular distance: %f %s\n", max_angle, (angle_units) ? "arcmin" : "degrees");
    printf("Bins per dec: %i\n", bins_per_dec);
    printf("Total bins  : %i\n", *nbins);

    printf("Bin boundaries:\n");
    printf("#  degrees          arcminutes        dot_product\n");
    for (k = 0; k < (*nbins)+1; k++)
    {
        double bb  = pow(10, log10(min_angle) + k*1.0/bins_per_dec);
        binb[k] = cos(bb / ((angle_units) ? 60.0 : 1.0) * d2r);

        printf("%i %.15f %.15f ", k, (angle_units) ? bb/60.0f : bb, (angle_units) ? bb : bb * 60);
        printf("%.15f\n", binb[k]);
    }
    printf("\n");

    return binb;
}


void doCompute(struct accelerator *myacc, int co_proc,
               struct cartesian *data1, int n1, 
               struct cartesian *data2, int n2,
               int doSelf, 
               long long **data_bins, int nbins, double *binb, int njk)
{
    int i, j, segsize;

    if (co_proc == CO_PROC_FPGA) segsize = 32768; // to avoid overflow
    else if (co_proc == CO_PROC_CPU) segsize = 1024; // benefit from CPU cache
    else segsize = n1; // don't care 

    if (doSelf) // just to make sure
    {
        data2 = data1;
        n2 = n1;
    }

    // generate all jobs descriptors
    for (i = 0; i < n1; i += segsize)
    {
        int n1s = (segsize < n1-i) ? segsize : n1-i; 

        for (j = ((doSelf) ? i : 0); j < n2; j += segsize)
        {
            int n2s = (segsize < n2-j) ? segsize : n2-j; 

            if (co_proc == CO_PROC_FPGA)
                doComputeFPGA(myacc, &data1[i], n1s, &data2[j], n2s, (doSelf && i==j), data_bins, nbins, binb, njk);
            else
                doComputeWATERFALL(&data1[i], n1s, &data2[j], n2s, (doSelf && i==j), data_bins, nbins, binb, njk);
        }
    }
}



